@@ -1,7 +1,7 @@
 /*
 	mpg123lib_intern: Common non-public stuff for libmpg123
 
-	copyright 1995-2008 by the mpg123 project - free software under the terms of the LGPL 2.1
+	copyright 1995-2023 by the mpg123 project - free software under the terms of the LGPL 2.1
 	see COPYING and AUTHORS files in distribution or http://mpg123.org
 
 	derived from the old mpg123.h
@@ -14,37 +14,15 @@
 #define MPG123_ENCODINGS 12
 
 #include "config.h" /* Load this before _anything_ */
-#include "intsym.h" /* Prefixing of internal symbols that still are public in a static lib. */
 
-/* ABI conformance for other compilers.
-   mpg123 needs 16byte-aligned stack for SSE and friends.
-   gcc provides that, but others don't necessarily. */
-#ifdef ABI_ALIGN_FUN
-#ifndef attribute_align_arg
-#if defined(__GNUC__) && (__GNUC__ > 4 || __GNUC__ == 4 && __GNUC_MINOR__>1)
-#    define attribute_align_arg __attribute__((force_align_arg_pointer))
-/* The gcc that can align the stack does not need the check... nor does it work with gcc 4.3+, anyway. */
-#else
+#include "../common/abi_align.h"
 
-#    define attribute_align_arg
-/* Other compilers get code to catch misaligned stack.
-   Well, except Sun Studio, which accepts the aligned attribute but does not honor it. */
-#if !defined(__SUNPRO_C)
-#    define NEED_ALIGNCHECK
-#endif
+#include "../compat/compat.h"
 
-#endif
-#endif
-#else
-#define attribute_align_arg
-/* We won't try the align check... */
-#endif
-
-/* export DLL symbols */
-#if defined(WIN32) && defined(DYNAMIC_BUILD)
-#define BUILD_MPG123_DLL
-#endif
-#include "compat.h"
+// Only portable API plays a role in the library itself, outside of lfs_wrap.c.
+// Also, we need to ensure no suffix renaming for the primary implementations.
+#define MPG123_PORTABLE_API
+#define MPG123_ENUM_API
 #include "mpg123.h"
 
 #define SKIP_JUNK 1
@@ -60,37 +38,57 @@
 #define memmove(dst,src,size) bcopy(src,dst,size)
 #endif
 
-/* some stuff has to go back to mpg123.h */
-#ifdef REAL_IS_FLOAT
-#  define real float
-#  define REAL_SCANF "%f"
-#  define REAL_PRINTF "%f"
-#elif defined(REAL_IS_LONG_DOUBLE)
-#  define real long double
-#  define REAL_SCANF "%Lf"
-#  define REAL_PRINTF "%Lf"
-#elif defined(REAL_IS_FIXED)
-/* Disable some output formats for fixed point decoder... */
+// The real type is either 32 bit float or integer. We do not attempt
+// double precision for MPEG audio decoding (or any audio storage/compression
+// format, even).
+// The calctables tool works in double precision float, but converts to fixed
+// point 'real' on output in addition to floating point.
 
-# define real long
+// The base type used in the computation code.
+// With C11 we  could rely on tgmath for sin() to use the right precision.
+#ifdef CALCTABLES
+#	define clreal double
+#	define COS    cos
+#	define SIN    sin
+#	define TAN    tan
+#	define POW    pow
+#else
+#	define clreal float
+#	define COS    cosf
+#	define SIN    sinf
+#	define TAN    tanf
+#	define POW    powf
+#endif
+
+#if defined(REAL_IS_FLOAT) && !defined(CALCTABLES)
+
+#define real float
+
+#elif defined(REAL_IS_FIXED) || defined(CALCTABLES)
+
+#ifdef RUNTIME_TABLES
+#error "Runtime tables are only for floating point decoders."
+#endif
+
+# define real  int32_t
+# define dreal int64_t
 
 /*
   for fixed-point decoders, use pre-calculated tables to avoid expensive floating-point maths
   undef this macro for run-time calculation
 */
-#define PRECALC_TABLES
 
 # define REAL_RADIX				24
 # define REAL_FACTOR			16777216.0
 
-static inline long double_to_long_rounded(double x, double scalefac)
+static inline int32_t double_to_long_rounded(double x, double scalefac)
 {
 	x *= scalefac;
 	x += (x > 0) ? 0.5 : -0.5;
-	return (long)x;
+	return (int32_t)x;
 }
 
-static inline long scale_rounded(long x, int shift)
+static inline int32_t scale_rounded(int32_t x, int shift)
 {
 	x += (x >> 31);
 	x >>= (shift - 1);
@@ -190,11 +188,14 @@ static inline long scale_rounded(long x, int shift)
 #  endif
 # endif
 
-/* I just changed the (int) to (long) there... seemed right. */
+/* I just changed the (int) to (real) there... seemed right. */
 # define DOUBLE_TO_REAL(x)					(double_to_long_rounded(x, REAL_FACTOR))
-# define DOUBLE_TO_REAL_15(x)				(double_to_long_rounded(x, 32768.0))
-# define DOUBLE_TO_REAL_POW43(x)			(double_to_long_rounded(x, 8192.0))
-# define DOUBLE_TO_REAL_SCALE_LAYER12(x)	(double_to_long_rounded(x, 1073741824.0))
+# define SCALE_15								32768.0
+# define DOUBLE_TO_REAL_15(x)				(double_to_long_rounded(x, SCALE_15))
+# define SCALE_POW43							8192.0
+# define DOUBLE_TO_REAL_POW43(x)			(double_to_long_rounded(x, SCALE_POW43))
+# define SCALE_LAYER12                  1073741824.0
+# define DOUBLE_TO_REAL_SCALE_LAYER12(x)	(double_to_long_rounded(x, SCALE_LAYER12))
 # define DOUBLE_TO_REAL_SCALE_LAYER3(x, y)	(double_to_long_rounded(x, pow(2.0,gainpow2_scale[y])))
 # define REAL_TO_DOUBLE(x)					((double)(x) / REAL_FACTOR)
 # ifdef REAL_MUL_ASM
@@ -202,17 +203,17 @@ static inline long scale_rounded(long x, int shift)
 #  define REAL_MUL_15(x, y)					REAL_MUL_ASM(x, y, 15)
 #  define REAL_MUL_SCALE_LAYER12(x, y)		REAL_MUL_ASM(x, y, 15 + 30 - REAL_RADIX)
 # else
-#  define REAL_MUL(x, y)					(((long long)(x) * (long long)(y)) >> REAL_RADIX)
-#  define REAL_MUL_15(x, y)					(((long long)(x) * (long long)(y)) >> 15)
-#  define REAL_MUL_SCALE_LAYER12(x, y)		(((long long)(x) * (long long)(y)) >> (15 + 30 - REAL_RADIX))
+#  define REAL_MUL(x, y)					(((dreal)(x) * (dreal)(y)) >> REAL_RADIX)
+#  define REAL_MUL_15(x, y)					(((dreal)(x) * (dreal)(y)) >> 15)
+#  define REAL_MUL_SCALE_LAYER12(x, y)		(((dreal)(x) * (dreal)(y)) >> (15 + 30 - REAL_RADIX))
 # endif
 # ifdef REAL_MUL_SCALE_LAYER3_ASM
 #  define REAL_MUL_SCALE_LAYER3(x, y, z)	REAL_MUL_SCALE_LAYER3_ASM(x, y, 13 + gainpow2_scale[z] - REAL_RADIX)
 # else
-#  define REAL_MUL_SCALE_LAYER3(x, y, z)	(((long long)(x) * (long long)(y)) >> (13 + gainpow2_scale[z] - REAL_RADIX))
+#  define REAL_MUL_SCALE_LAYER3(x, y, z)	(((dreal)(x) * (dreal)(y)) >> (13 + gainpow2_scale[z] - REAL_RADIX))
 # endif
-# define REAL_SCALE_LAYER12(x)				((long)((x) >> (30 - REAL_RADIX)))
-# define REAL_SCALE_LAYER3(x, y)			((long)((x) >> (gainpow2_scale[y] - REAL_RADIX)))
+# define REAL_SCALE_LAYER12(x)				((real)((x) >> (30 - REAL_RADIX)))
+# define REAL_SCALE_LAYER3(x, y)			((real)((x) >> (gainpow2_scale[y] - REAL_RADIX)))
 # ifdef ACCURATE_ROUNDING
 #  define REAL_MUL_SYNTH(x, y)				REAL_MUL(x, y)
 #  define REAL_SCALE_DCT64(x)				(x)
@@ -222,19 +223,11 @@ static inline long scale_rounded(long x, int shift)
 #  define REAL_SCALE_DCT64(x)				((x) >> 8)
 #  define REAL_SCALE_WINDOW(x)				scale_rounded(x, 16)
 # endif
-#  define REAL_SCANF "%ld"
-#  define REAL_PRINTF "%ld"
 
 #else
-#  define real double
-#  define REAL_SCANF "%lf"
-#  define REAL_PRINTF "%f"
-#endif
 
-#ifndef REAL_IS_FIXED
-# if (defined SIZEOF_INT32_T) && (SIZEOF_INT32_T != 4)
-#  error "Bad 32bit types!!!"
-# endif
+#error "Simple float or fixed-point, nothing else makes sense."
+
 #endif
 
 #ifndef DOUBLE_TO_REAL
@@ -285,7 +278,7 @@ static inline long scale_rounded(long x, int shift)
    now: factor on minimum frame buffer size (which takes upsampling into account) */
 #define		AUDIOBUFSIZE		2
 
-#include "true.h"
+#include "../common/true.h"
 
 #define         MAX_NAME_SIZE           81
 #define         SBLIMIT                 32
@@ -321,13 +314,17 @@ static inline long scale_rounded(long x, int shift)
 #define VERBOSE4 (NOQUIET && fr->p.verbose > 3)
 #define PVERB(mp, level) (!((mp)->flags & MPG123_QUIET) && (mp)->verbose >= (level))
 
-int decode_update(mpg123_handle *mh);
+int INT123_decode_update(mpg123_handle *mh);
 /* residing in format.c  */
-off_t samples_to_storage(mpg123_handle *fr , off_t s);
-off_t samples_to_bytes(mpg123_handle *fr , off_t s);
-off_t bytes_to_samples(mpg123_handle *fr , off_t b);
+int64_t INT123_decoder_synth_bytes(mpg123_handle *fr , int64_t s);
+int64_t INT123_samples_to_bytes(mpg123_handle *fr , int64_t s);
+int64_t INT123_bytes_to_samples(mpg123_handle *fr , int64_t b);
+int64_t INT123_outblock_bytes(mpg123_handle *fr, int64_t s);
 /* Postprocessing format conversion of freshly decoded buffer. */
-void postprocess_buffer(mpg123_handle *fr);
+void INT123_postprocess_buffer(mpg123_handle *fr);
+
+int INT123_open_fixed_pre(mpg123_handle *mh, int channels, int encoding);
+int INT123_open_fixed_post(mpg123_handle *mh, int channels, int encoding);
 
 /* If networking is enabled and we really mean internal networking, the timeout_read function is available. */
 #if defined (NETWORK) && !defined (WANT_WIN32_SOCKETS)
@@ -335,4 +332,16 @@ void postprocess_buffer(mpg123_handle *fr);
 #define TIMEOUT_READ
 #endif
 
+// Change a given linear factor by the given dB value, bounded
+// to +/- 60 dB.
+static inline double dbchange(double base_factor, double db)
+{
+	double nscale = base_factor * pow(10, db/20);
+	if(nscale < 0.001) // -60 dB
+		nscale = 0.001;
+	if(nscale > 1000)
+		nscale = 1000; // +60 dB
+	return nscale;
+}
+
 #endif
