@@ -1,6 +1,6 @@
 #include "config.h"
 #include "mpg123app.h"
-#include "debug.h"
+#include "common/debug.h"
 
 #ifdef WANT_WIN32_UNICODE
 
@@ -37,7 +37,7 @@ int win32_cmdline_utf8(int * argc, char *** argv)
 
 	for(argcounter = 0; argcounter < *argc; argcounter++)
 	{
-		win32_wide_utf8(argv_wide[argcounter], &argvptr, NULL);
+		INT123_win32_wide_utf8(argv_wide[argcounter], &argvptr, NULL);
 		(*argv)[argcounter] = argvptr;
 	}
 	return 0;
@@ -51,6 +51,7 @@ void win32_cmdline_free(int argc, char **argv)
 }
 #endif /* WANT_WIN32_UNICODE */
 
+#ifdef _WIN32
 void win32_set_priority (const int arg)
 {
 	DWORD proc_result = 0;
@@ -74,6 +75,7 @@ void win32_set_priority (const int arg)
 		else fprintf(stderr,"GetCurrentProcess failed\n");
 	}
 }
+#endif
 
 #ifdef WANT_WIN32_FIFO
 static HANDLE fifohandle;
@@ -91,10 +93,9 @@ VOID CALLBACK ReadComplete(
 	return;
 }
 
-ssize_t win32_fifo_read(void *buf, size_t nbyte)
+mpg123_ssize_t win32_fifo_read(void *buf, size_t nbyte)
 {
 	int check;
-	DWORD re;
 	DWORD readbuff;
 	DWORD available;
 	debug1("Reading pipe handle %p", fifohandle);
@@ -128,7 +129,7 @@ DWORD win32_fifo_read_peek(struct timeval *tv)
 	if(!fifohandle) return 0;
 		PeekNamedPipe(fifohandle, NULL, 0, NULL, &ret, NULL);
 	err =  GetLastError();
-	debug1("Waiting %d msec for pipe to be ready", timer);
+	debug1("Waiting %ld msec for pipe to be ready", timer);
 	debug1("GetLastError was %ld", err);
 	if(err == ERROR_BROKEN_PIPE)
 	{
@@ -144,7 +145,7 @@ DWORD win32_fifo_read_peek(struct timeval *tv)
 		ConnectNamedPipe(fifohandle,&ov1);
 		WaitForSingleObjectEx(fifohandle,timer,TRUE);
 	}
-	debug2("peek %d bytes, error %d",ret, err);
+	debug2("peek %ld bytes, error %ld",ret, err);
 	return ret;
 }
 
@@ -165,7 +166,7 @@ int win32_fifo_mkfifo(const char *path)
 	win32_fifo_close();
 #ifdef WANT_WIN32_UNICODE
 	wchar_t *str;
-	if(win32_utf8_wide(path,&str,NULL) == 0)
+	if(INT123_win32_utf8_wide(path,&str,NULL) == 0)
 	{
 		fprintf(stderr,"Cannot get FIFO name, likely out of memory\n");
 		return -1;
